//
//  imageProcessing.c
//  UAV Imaging
//
//  Created by Shen Yue on 12-02-27.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#include <stdio.h>
#include <cv.h>
#include "imageProcessing.h"
#include <highgui.h>
#include "autopilot.h"
#include "configuration.h"
#include "utility.h"
#include <math.h>
#include "recognition.h"

// to do: change cv scalar to int[3]
typedef struct grid {
    CvScalar edge[8];
    CvScalar combinedEdge[8];
    CvScalar mean;
    int edgeOrientation;
    int potentialTarget;
} GRID;

GRID *** initializeGridData(IplImage* image,int gridSize);
IplImage* loadImage(int imageNumber);
float loadPixelsPerMeter(int imageNumber);
int processImage(IplImage* image,float pixelsPerMeter,int imageNumber,int compressionQuality);
void releaseFilterData(INT_POINTER_QUEUE *filterDataReleaseList);
void compressImage(IplImage* image,int imageNumber,int targetCount,int compressionQuality);
void averageGridEdgeColorVertical(IplImage* image,GRID ***grid,int length,int i,int j,int x,int y,int edge);
void averageGridEdgeColorHorizontal(IplImage* image,GRID ***grid,int length,int i,int j,int x,int y,int edge);
void findEdge(GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *gridQueue);
float calculateDeltaE(CvScalar color1, CvScalar color2);
void filterFalseEdge(GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *gridQueue);
CvScalar averageColor(CvScalar color1,CvScalar color2);
void freeGrid(GRID ***grid,int gridWidth,int gridHeight);
int sizeFilter(GRID ***grid,ARRAY_TWO_INT_QUEUE *gridQueue,int gridWidth, int gridHeight,int imageNumber,int halfGridSize,IplImage* image);
void connectGrid(int *minX,int *maxX,int *minY,int *maxY,GRID ***grid,int gridWidth, int gridHeight,int *gridCount,ARRAY_TWO_INT_QUEUE *buffer,IplImage* object,int objectNumber);
void drawGrid(IplImage* image,GRID ***grid,int gridSize);
void RGB2LAB(float r,float g, float b,CvScalar *lab);
void unconnectedGridFilter(GRID ***grid,ARRAY_TWO_INT_QUEUE *gridQueue,int gridWidth, int gridHeight);
int unconnectedGrid(int x,int y,int gridWidth, int gridHeight,GRID ***grid,ARRAY_TWO_INT_QUEUE *buffer);
int connectedEdge(GRID *grid1,GRID *grid2);
int varifyTarget(int minX,int minY,int maxX,int maxY,int gridCount,IplImage* object,int objectNumber,int imageNumber,int halfGridSize,IplImage* image);
void writeTargetInfo(int imageNumber,int minX,int minY,int maxX,int maxY,char* name);
void filterBackgroundEdge(GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *gridQueue);
int connectBackgroundGrid(int x,int y,GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *buffer);
int connectedEdgeStrict(GRID *grid1,GRID *grid2);
void cleanupGridQueue(ARRAY_TWO_INT_QUEUE *gridQueue,GRID ***grid);

int processControl(int imageNumber,int compressionQuality)
{
    int targetCount;
    IplImage* image=loadImage(imageNumber);
    if (image!=NULL)
    {
        float pixelsPerMeter=loadPixelsPerMeter(imageNumber);
        targetCount=processImage(image, pixelsPerMeter,imageNumber,compressionQuality);
        cvReleaseImage(&image);
        //printf("processed:%d\n",imageNumber);
    }
    return targetCount;
}

IplImage* loadImage(int imageNumber)
{
    IplImage* imageBayer;
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%d.%s",FILE_PATH_CAPTURED,imageNumber,CAPTURED_IMAGE_FORMAT);
    imageBayer = cvLoadImage(path,CV_LOAD_IMAGE_GRAYSCALE);
    if (imageBayer==NULL)
    {
        printf("Error loading image %s\n",path);
    }
    IplImage* image;
    image=cvCreateImage(cvGetSize(imageBayer) , IPL_DEPTH_8U ,3);
    cvCvtColor(imageBayer, image, CV_BayerGB2RGB);
    cvReleaseImage(&imageBayer);
    return image;
}

float loadPixelsPerMeter(int imageNumber)
{
    float altitude;
    char inputBuffer[MAX_FILE_INPUT_LINE_SIZE];
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%d.%s",FILE_PATH_INFO,imageNumber,INFO_FORMAT);
    FILE* fp = fopen(path, "r");
    if (fp==NULL)
    {
        printf("Error reading info %s\n",path);
        return 100; //for testing
    }
    char* ret= fgets(inputBuffer,MAX_FILE_INPUT_LINE_SIZE,fp);
    if (ret==NULL)
    {
        printf("Error reading info %s\n",path);
    }
    sscanf(inputBuffer, "altitude: %f",&altitude);
    fclose(fp);
    return getPixelsPerMeter(altitude);
}

int processImage(IplImage* image,float pixelsPerMeter,int imageNumber,int compressionQuality)
{
    int gridSize=IMAGE_PROCESSING_GRID_SIZE;
    int halfGridSize=gridSize/2;
    int gridWidth=(image->width-halfGridSize-1)/halfGridSize;
    int gridHeight=(image->height-halfGridSize-1)/halfGridSize;
    int targetCount=0;
    
    TWO_D_INT_POINTER_ARRAY *filterData=mallocTwoDIntPointerArray(gridWidth, gridHeight);
    INT_POINTER_QUEUE *filterDataReleaseList= mallocIntPointerQueue();
    ARRAY_TWO_INT_QUEUE *gridQueue=mallocArrayTwoIntQueue(gridWidth*gridHeight);
    
    GRID ***grid=initializeGridData(image,gridSize);
    
    findEdge(grid,gridWidth, gridHeight,gridQueue);
    
    filterBackgroundEdge(grid, gridWidth, gridHeight,gridQueue);
    
    filterFalseEdge(grid, gridWidth, gridHeight,gridQueue);
    
    unconnectedGridFilter(grid, gridQueue, gridWidth, gridHeight);
    
    //drawGrid(image, grid, gridSize);
    
    sizeFilter(grid, gridQueue, gridWidth, gridHeight,imageNumber,halfGridSize,image);
    
    compressImage(image, imageNumber, targetCount,compressionQuality);
    
    freeArrayTwoIntQueue(gridQueue);
    freeTwoDIntPointerArray(filterData);
    releaseFilterData(filterDataReleaseList);
    freeGrid(grid,gridWidth,gridHeight);
    return targetCount;
}

void compressImage(IplImage* image,int imageNumber,int targetCount,int compressionQuality)
{
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%d.%s",FILE_PATH_PROCESSED,imageNumber,PROCESSED_IMAGE_FORMAT);
    int params[3];
    params[0] = CV_IMWRITE_JPEG_QUALITY;
    if (targetCount>0)
    {
        params[1]=compressionQuality*MIN_COMPRESSION_QUALITY_FOR_TARGET_IMAGE/MIN_COMPRESSION_QUALITY_FOR_OTHER_IMAGE;
        if (params[1]>MAX_COMPRESSION_QUALITY)
        {
            params[1]=MAX_COMPRESSION_QUALITY;
        }
    }
    else
    {
        params[1] = compressionQuality;
    }
    params[2] = 0;
    cvSaveImage(path, image,params);
}

void releaseFilterData(INT_POINTER_QUEUE *filterDataReleaseList)
{
    int *pointer;
    int ret=removeFromIntPointerQueue(filterDataReleaseList, &pointer);
    while (ret == 1)
    {
        free(pointer);
        ret=removeFromIntPointerQueue(filterDataReleaseList, &pointer);
    }
    free(filterDataReleaseList);
}

GRID *** initializeGridData(IplImage* image,int gridSize)
{
    int i,j,k;
    int x,y;
    int width=image->width;
    int height=image->height;
    int halfGridSize=gridSize/2;
    int gridWidth=(width-halfGridSize-1)/halfGridSize;
    int gridHeight=(height-halfGridSize-1)/halfGridSize;
    
    GRID ***grid=malloc(gridWidth*sizeof(GRID**));
    
    for( i=0;i<gridWidth;i++)
    {
        grid[i]=malloc(gridHeight*sizeof(GRID*));
        for (j=0; j<gridHeight; j++) 
        {
            grid[i][j]=malloc(sizeof(GRID));
            grid[i][j]->edgeOrientation=-1;
            grid[i][j]->potentialTarget=0;
        }
    }
    
    //Horizontal
    
    y=0;
    for (j=0;j<gridSize;j+=halfGridSize)
    {
        averageGridEdgeColorHorizontal(image,grid,halfGridSize,1,j,0,y,0);
        averageGridEdgeColorHorizontal(image,grid,halfGridSize,1+halfGridSize,j,0,y,1);
        
        grid[0][y]->combinedEdge[0]=averageColor(grid[0][y]->edge[0], grid[0][y]->edge[1]);
        
        averageGridEdgeColorHorizontal(image,grid,halfGridSize,1,j+gridSize,0,y,5);
        averageGridEdgeColorHorizontal(image,grid,halfGridSize,1+halfGridSize,j+gridSize,0,y,4);
        
        grid[0][y]->combinedEdge[4]=averageColor(grid[0][y]->edge[5], grid[0][y]->edge[4]);
        
        x=1;
        for (i=1+halfGridSize;i<width-gridSize+1;i+=halfGridSize)
        {
            grid[x][y]->edge[0]=grid[x-1][y]->edge[1];
            averageGridEdgeColorHorizontal(image,grid,halfGridSize,i+halfGridSize,j,x,y,1);
            
            grid[x][y]->combinedEdge[0]=averageColor(grid[x][y]->edge[0], grid[x][y]->edge[1]);
            
            grid[x][y]->edge[5]=grid[x-1][y]->edge[4];
            averageGridEdgeColorHorizontal(image,grid,halfGridSize,i+halfGridSize,j+gridSize,x,y,4);
            
            grid[x][y]->combinedEdge[4]=averageColor(grid[x][y]->edge[5], grid[x][y]->edge[4]);
            
            x++;
        }
        y++;
    }
    
    y=2;
    for (j=gridSize;j<height-gridSize;j+=halfGridSize)
    {
        grid[0][y]->edge[0]=grid[0][y-2]->edge[5];
        grid[0][y]->edge[1]=grid[0][y-2]->edge[4];
        
        grid[0][y]->combinedEdge[0]=grid[0][y-2]->combinedEdge[4];
        
        averageGridEdgeColorHorizontal(image,grid,halfGridSize,1,j+gridSize,0,y,5);
        averageGridEdgeColorHorizontal(image,grid,halfGridSize,1+halfGridSize,j+gridSize,0,y,4);
        
        grid[0][y]->combinedEdge[4]=averageColor(grid[0][y]->edge[5], grid[0][y]->edge[4]);
        
        x=1;
        for (i=1+halfGridSize;i<width-gridSize+1;i+=halfGridSize)
        {
            grid[x][y]->edge[0]=grid[x][y-2]->edge[5];
            grid[x][y]->edge[1]=grid[x][y-2]->edge[4];
            
            grid[x][y]->combinedEdge[0]=grid[x][y-2]->combinedEdge[4];
            
            grid[x][y]->edge[5]=grid[x-1][y]->edge[4];
            averageGridEdgeColorHorizontal(image,grid,halfGridSize,i+halfGridSize,j+gridSize,x,y,4);
            
            grid[x][y]->combinedEdge[4]=averageColor(grid[x][y]->edge[5], grid[x][y]->edge[4]);
            
            x++;
        }
        y++;
    }
    
    //Vertical
    
    x=0;
    for (i=0;i<gridSize;i+=halfGridSize)
    {
        averageGridEdgeColorVertical(image,grid,halfGridSize,i,1,x,0,7);
        averageGridEdgeColorVertical(image,grid,halfGridSize,i,1+halfGridSize,x,0,6);
        
        grid[x][0]->combinedEdge[6]=averageColor(grid[x][0]->edge[7], grid[x][0]->edge[6]);
        
        averageGridEdgeColorVertical(image,grid,halfGridSize,i+gridSize,1,x,0,2);
        averageGridEdgeColorVertical(image,grid,halfGridSize,i+gridSize,1+halfGridSize,x,0,3);
        
        grid[x][0]->combinedEdge[2]=averageColor(grid[x][0]->edge[2], grid[x][0]->edge[3]);
        
        y=1;
        for (j=1+halfGridSize;j<height-gridSize+1;j+=halfGridSize)
        {
            grid[x][y]->edge[7]=grid[x][y-1]->edge[6];
            averageGridEdgeColorVertical(image,grid,halfGridSize,i,j+halfGridSize,x,y,6);
            
            grid[x][y]->combinedEdge[6]=averageColor(grid[x][y]->edge[7], grid[x][y]->edge[6]);
            
            grid[x][y]->edge[2]=grid[x][y-1]->edge[3];
            averageGridEdgeColorVertical(image,grid,halfGridSize,i+gridSize,j+halfGridSize,x,y,3);
            
            grid[x][y]->combinedEdge[2]=averageColor(grid[x][y]->edge[2], grid[x][y]->edge[3]);
            
            y++;
        }
        x++;
    }
    
    x=2;
    for (i=gridSize;i<width-gridSize;i+=halfGridSize)
    {
        grid[x][0]->edge[7]=grid[x-2][0]->edge[2];
        grid[x][0]->edge[6]=grid[x-2][0]->edge[3];
        
        grid[x][0]->combinedEdge[6]=grid[x-2][0]->combinedEdge[2];
        
        averageGridEdgeColorVertical(image,grid,halfGridSize,i+gridSize,1,x,0,2);
        averageGridEdgeColorVertical(image,grid,halfGridSize,i+gridSize,1+halfGridSize,x,0,3);
        
        grid[x][0]->combinedEdge[2]=averageColor(grid[x][0]->edge[2], grid[x][0]->edge[3]);
        
        y=1;
        for (j=1+halfGridSize;j<height-gridSize+1;j+=halfGridSize)
        {
            grid[x][y]->edge[7]=grid[x-2][y]->edge[2];
            grid[x][y]->edge[6]=grid[x-2][y]->edge[3];
            
            grid[x][y]->combinedEdge[6]=grid[x-2][y]->combinedEdge[2];
            
            grid[x][y]->edge[2]=grid[x][y-1]->edge[3];
            averageGridEdgeColorVertical(image,grid,halfGridSize,i+gridSize,j+halfGridSize,x,y,3);
            
            grid[x][y]->combinedEdge[2]=averageColor(grid[x][y]->edge[2], grid[x][y]->edge[3]);
            
            y++;
        }
        x++;
    }
    
    float totalL;
    float totalA;
    float totalB;
    
    for (i=0;i<gridWidth;i++)
    {
        for (j=0;j<gridHeight;j++)
        {
            //combined edge
            for (k=1; k<8; k+=2) 
            {
                grid[i][j]->combinedEdge[k]=averageColor(grid[i][j]->edge[k], grid[i][j]->edge[(k+1)%8]);
            }
            
            //mean
            totalL=0;
            totalA=0;
            totalB=0;
            
            for(k=0;k<8;k+=2)
            {
                totalL+=grid[i][j]->combinedEdge[k].val[0];
                totalA+=grid[i][j]->combinedEdge[k].val[1];
                totalB+=grid[i][j]->combinedEdge[k].val[2];
            }
            grid[i][j]->mean.val[0]=totalL/4;
            grid[i][j]->mean.val[1]=totalA/4;
            grid[i][j]->mean.val[2]=totalB/4;  
        }
    }
    return grid;
}

void averageGridEdgeColorVertical(IplImage* image,GRID ***grid,int length,int i,int j,int x,int y,int edge)
{
    int step       = image->widthStep/sizeof(uchar);
    int channels   = image->nChannels;
    uchar* data    = (uchar *)image->imageData;
    
    int k;
    float totalL=0;
    float totalA=0;
    float totalB=0;
    
    for (k=j; k<j+length-1; k++) {
        totalL+=data[k*step+i*channels+0];
        totalA+=data[k*step+i*channels+1];
        totalB+=data[k*step+i*channels+2];
        //cvLine(image, cvPoint(i, k), cvPoint(i, k), cvScalar(0, 0, 0, 0), 1, 1, 0);
    }
    RGB2LAB(totalL/length,totalA/length,totalB/length,  &grid[x][y]->edge[edge]);
}

void averageGridEdgeColorHorizontal(IplImage* image,GRID ***grid,int length,int i,int j,int x,int y,int edge)
{
    int step       = image->widthStep/sizeof(uchar);
    int channels   = image->nChannels;
    uchar* data    = (uchar *)image->imageData;
    int k;
    float totalL=0;
    float totalA=0;
    float totalB=0;
    
    for (k=i; k<i+length-1; k++) {
        totalL+=data[j*step+k*channels+0];
        totalA+=data[j*step+k*channels+1];
        totalB+=data[j*step+k*channels+2];
        //cvLine(image, cvPoint(k,j), cvPoint(k, j), cvScalar(0, 0, 0, 0), 1, 1, 0);
        
    }
    RGB2LAB(totalL/length,totalA/length,totalB/length,  &grid[x][y]->edge[edge]);
}

void RGB2LAB(float r,float g, float b,CvScalar *lab)
{
    r/=255.0;
    g/=255.0;
    b/=255.0;
    if (r>0.04045) 
    {
        r=powf((r+0.055)/1.055,2.4);
    }
    else
    {
        r/=12.92;
    }
    if (g>0.04045)
    {
        g=powf((g+0.055)/1.055,2.4);
    }
    else
    {
        g/=12.92;
    }
    if (b>0.04045) 
    {
        b=powf((b+0.055)/1.055,2.4);
    }
    else 
    {
        b/=12.92;
    }
    r*=100.0;
    g*=100.0;
    b*=100.0;
    float x=r*0.4124+g*0.3576+b*0.1805;
    float y=r*0.2126+g*0.7152+b*0.0722;
    float z=r*0.0193+g*0.1192+b*0.9505;
    //xyz to lab
    x/=95.047;
    y/=100.000;
    z/=108.883;
    if (x>0.008856) 
    {
        x=powf(x,0.3333);
    }
    else
    {
        x=(7.787*x)+0.1379;
    }
    if (y>0.008856)
    {
        y=powf(y,0.3333);
    }
    else
    {
        y=(7.787*y)+0.1379;
    }
    if (z>0.008856)
    {
        z=powf(z,0.3333);
    }
    else
    {
        z=(7.787*z)+0.1379;
    }    
    lab->val[0]=(116.0*y)-16.0;
    lab->val[1]=500.0*(x-y);
    lab->val[2]=200.0*(y-z);
}

float calculateDeltaE(CvScalar color1, CvScalar color2)
{
    float l1=color1.val[0];
    float l2=color2.val[0];
    float a1=color1.val[1];
    float a2=color2.val[1];
    float b1=color1.val[2];
    float b2=color2.val[2];
    float deltaL=l1-l2;
    float c1=sqrtf(a1*a1+b1*b1);
    float c2=sqrtf(a2*a2+b2*b2);
    float deltaC=c1-c2;
    float deltaA=a1-a2;
    float deltaB=b1-b2;
    float deltaHSquared=deltaA*deltaA+deltaB*deltaB-deltaC*deltaC;
    float deltaH;
    if (deltaHSquared>0)
    {
        deltaH=sqrtf(deltaHSquared);
    }
    else
    {
        deltaH=0;
    }
    float q1=deltaL/DELTA_E_K_L;
    float q2=deltaC/(1+DELTA_E_K_1*c1);
    float q3=deltaH/(1+DELTA_E_K_2*c2);
    float deltaE=sqrtf(q1*q1+q2*q2+q3*q3);
    return deltaE;
}

void findEdge(GRID ***grid,int gridWidth, int gridHeight, ARRAY_TWO_INT_QUEUE *gridQueue)
{
    int i,j,k;
    float deltaE;
    float maxDeltaE;
    for (i=0; i<gridWidth; i++) {
        for (j=0;j<gridHeight; j++) {
            maxDeltaE=0;
            for(k=0;k<4;k++)
            {
                deltaE=calculateDeltaE(grid[i][j]->combinedEdge[k],grid[i][j]->combinedEdge[k+4]);
                if (deltaE>DELTA_E_EDGE_GRID_THRESHOLD && deltaE>maxDeltaE) 
                {
                    grid[i][j]->edgeOrientation=k;
                    maxDeltaE=deltaE;
                }
            }
            if (grid[i][j]->edgeOrientation>=0)
            {
                grid[i][j]->potentialTarget=1;
                addToArrayTwoIntQueue(gridQueue, i, j);
            }
        }
    }
}

CvScalar averageColor(CvScalar color1,CvScalar color2)
{
    return cvScalar((color1.val[0]+color2.val[0])/2, (color1.val[1]+color2.val[1])/2, (color1.val[2]+color2.val[2])/2, 0);
}

void drawGrid(IplImage* image,GRID ***grid,int gridSize)
{
    int halfGridSize=gridSize/2;
    int fourthGridSize=gridSize/4;
    int width=image->width;
    int height=image->height;
    int i,j;
    int x,y;
    x=0;
    for (i=0;i<width-gridSize;i+=halfGridSize)
    {
        y=0;
        for (j=1;j<height-gridSize+1;j+=halfGridSize)
        {
            if (grid[x][y]->potentialTarget) 
            {
                cvRectangle(image, cvPoint(i+1, j+1), cvPoint(i+halfGridSize, j+halfGridSize), cvScalar(255, 255, 255, 0), 1, 1, 0);
                switch (grid[x][y]->edgeOrientation) {
                    case 0:
                        cvLine(image, cvPoint(i+1, j+fourthGridSize), cvPoint(i+halfGridSize, j+fourthGridSize), cvScalar(0, 0, 0, 0), 1, 1, 0);
                        break;
                    case 1:
                        cvLine(image, cvPoint(i+1, j+1), cvPoint(i+halfGridSize, j+halfGridSize), cvScalar(0, 0, 0, 0), 1, 1, 0);
                        break;
                    case 2:
                        cvLine(image, cvPoint(i+fourthGridSize, j+1), cvPoint(i+fourthGridSize, j+halfGridSize), cvScalar(0, 0, 0, 0), 1, 1, 0);
                        break;
                    case 3:
                        cvLine(image, cvPoint(i+1, j+halfGridSize), cvPoint(i+halfGridSize, j+1), cvScalar(0, 0, 0, 0), 1, 1, 0);
                        break;
                    default:
                        break;
                }
            }
            y++;
        }
        x++;
    }
    
    cvNamedWindow("window", CV_WINDOW_AUTOSIZE);
    cvShowImage("window", image);
    cvWaitKey(0);
}

void filterFalseEdge(GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *gridQueue)
{
    int i;
    int x,y;
    int x1,y1,x2,y2;
    int dx[4]={0,1,1,1};
    int dy[4]={-1,-1,0,1};
    int d;
    int count=gridQueue->count;
    
    for (i=0; i<count; i++)
    {
        removeFromArrayTwoIntQueue(gridQueue, &x, &y);
        d=grid[x][y]->edgeOrientation;
        x1=x-dx[d];
        y1=y-dy[d];
        if (x1>=0 && y1>=0 && x1<gridWidth && y1<gridHeight)
        {
            if (calculateDeltaE(grid[x][y]->combinedEdge[d], grid[x-dx[d]][y-dy[d]]->combinedEdge[d+4])<DELTA_E_FALSE_EDGE_THRESHOLD)
            {
                grid[x][y]->potentialTarget=0;
            }
            else
            {
                x2=x+dx[d];
                y2=y+dy[d];
                if (x2>=0 && y2>=0 && x2<gridWidth && y2<gridHeight)
                {
                    if (calculateDeltaE(grid[x][y]->combinedEdge[d], grid[x-dx[d]][y-dy[d]]->combinedEdge[d+4])<DELTA_E_FALSE_EDGE_THRESHOLD)
                    {
                        grid[x][y]->potentialTarget=0;
                    }
                    else
                    {
                        addToArrayTwoIntQueue(gridQueue, x, y);
                    }
                }
                else
                {
                    addToArrayTwoIntQueue(gridQueue, x, y);
                }
            }
        }
        else
        {
            addToArrayTwoIntQueue(gridQueue, x, y);
        }
    }
}

void unconnectedGridFilter(GRID ***grid,ARRAY_TWO_INT_QUEUE *gridQueue,int gridWidth, int gridHeight)
{
    int i;
    int x,y;
    int count=gridQueue->count;
    ARRAY_TWO_INT_QUEUE *buffer = mallocArrayTwoIntQueue(gridQueue->count);
    for (i=0; i<count; i++)
    {
        removeFromArrayTwoIntQueue(gridQueue, &x, &y);
        if (grid[x][y]->potentialTarget==1)
        {
            if(!unconnectedGrid(x,y,gridWidth,gridHeight,grid,buffer))
            {
                addToArrayTwoIntQueue(gridQueue, x, y);
            }
        }
    }
    cleanupGridQueue(gridQueue, grid);
    freeArrayTwoIntQueue(buffer);
}

int unconnectedGrid(int startingX,int startingY,int gridWidth, int gridHeight,GRID ***grid,ARRAY_TWO_INT_QUEUE *buffer)
{
    int i;
    int unconnected=0;
    int count=1;
    int xDirection[8]={-1,0,1,1,1,0,-1,-1};
    int yDirection[8]={-1,-1,-1,0,1,1,1,0};
    int x=startingX;
    int y=startingY;
    int currentX,currentY;
    int nextX,nextY;
    
    while (count==1) 
    {
        count=0;
        for (i=0; i<8; i++) 
        {
            currentX=x+xDirection[i];
            currentY=y+yDirection[i];
            if (currentX>=0 && currentX<gridWidth && currentY>=0 && currentY<gridHeight)
            {
                if (grid[currentX][currentY]->potentialTarget==1) 
                {
                    count++;
                    nextX=currentX;
                    nextY=currentY;
                    if (count>1) {
                        break;
                    }
                }
            }
        }
        if (count<2) 
        {
            grid[x][y]->potentialTarget=0;
            unconnected=1;
            if (count==1) 
            {
                x=nextX;
                y=nextY;
            }
        }
    }
    return unconnected;
}

int sizeFilter(GRID ***grid,ARRAY_TWO_INT_QUEUE *gridQueue,int gridWidth, int gridHeight,int imageNumber,int halfGridSize,IplImage* image)
{
    int i;
    int x,y;
    int count=gridQueue->count;
    int targetCount=0;
    int minX,minY,maxX,maxY;
    int gridCount;
    int objectNumber=1;
    ARRAY_TWO_INT_QUEUE *buffer = mallocArrayTwoIntQueue(gridQueue->count);
    
    IplImage* object;
    object=cvCreateImage(cvSize(gridWidth, gridHeight) , IPL_DEPTH_8U ,1);
    cvSet(object, cvScalar(0, 0, 0, 0), NULL);
    
    for (i=0; i<count; i++)
    {
        removeFromArrayTwoIntQueue(gridQueue, &x, &y);
        if (grid[x][y]->potentialTarget==1)
        {
            if (objectNumber>255)
            {
                objectNumber=1;
                cvSet(object, cvScalar(0, 0, 0, 0), NULL);
            }
            minX=x;
            minY=y;
            maxX=x;
            maxY=y;
            gridCount=0;
            connectGrid(&minX, &maxX, &minY, &maxY, grid, gridWidth, gridHeight, &gridCount,buffer,object,objectNumber);
            targetCount+=varifyTarget(minX,minY,maxX,maxY,gridCount,object,objectNumber,imageNumber,halfGridSize,image);
            objectNumber++;
        }
    }
    
    freeArrayTwoIntQueue(buffer);
    cvReleaseImage(&object);
    return targetCount;
}

void connectGrid(int *minX,int *maxX,int *minY,int *maxY,GRID ***grid,int gridWidth, int gridHeight,int *gridCount,ARRAY_TWO_INT_QUEUE *buffer,IplImage* object,int objectNumber)
{
    int step       = object->widthStep/sizeof(uchar);
    uchar* data    = (uchar *)object->imageData;
    
    int xDirection[8]={-1,0,1,1,1,0,-1,-1};
    int yDirection[8]={-1,-1,-1,0,1,1,1,0};
    grid[*minX][*minY]->potentialTarget=0;
    addToArrayTwoIntQueue(buffer, *minX, *minY);
    int currentX,currentY;
    int x,y;
    int i;
    
    while (removeFromArrayTwoIntQueue(buffer, &x, &y)) 
    {
        if (x<*minX)
        {
            *minX=x;
        }
        else if(x>*maxX)
        {
            *maxX=x;
        }
        if (y<*minY)
        {
            *minY=y;
        }
        else if(y>*maxY)
        {
            *maxY=y;
        }
        
        (*gridCount)++;
        
        for (i=0; i<8; i++) 
        {
            currentX=x+xDirection[i];
            currentY=y+yDirection[i];
            if (currentX>=0 && currentX<gridWidth && currentY>=0 && currentY<gridHeight)
            {
                if (grid[currentX][currentY]->potentialTarget==1 && connectedEdge(grid[x][y],grid[currentX][currentY]))
                {
                    grid[currentX][currentY]->potentialTarget=0;
                    addToArrayTwoIntQueue(buffer, currentX, currentY);
                    data[currentY*step+currentX]=objectNumber;
                }
            }
        }
    }
}

int connectedEdge(GRID *grid1,GRID *grid2)
{
    int orientation[2];
    orientation[0]=grid1->edgeOrientation;
    orientation[1]=grid2->edgeOrientation;
    float difference[2][2];
    
    difference[0][0]=calculateDeltaE(grid1->combinedEdge[orientation[0]], grid2->combinedEdge[orientation[0]]);
    difference[0][1]=calculateDeltaE(grid1->combinedEdge[orientation[1]+4], grid2->combinedEdge[orientation[1]+4]);
    if (difference[0][0]<DELTA_E_CONNECTED_EDGE_THRESHOLD && difference[0][1]<DELTA_E_CONNECTED_EDGE_THRESHOLD) 
    {
        return 1;
    }
    
    difference[1][0]=calculateDeltaE(grid1->combinedEdge[orientation[0]], grid2->combinedEdge[orientation[0]+4]);
    difference[1][1]=calculateDeltaE(grid1->combinedEdge[orientation[1]+4], grid2->combinedEdge[orientation[1]]);
    if (difference[1][0]<DELTA_E_CONNECTED_EDGE_THRESHOLD && difference[1][1]<DELTA_E_CONNECTED_EDGE_THRESHOLD)
    {
        return 1;
    }
    
    GRID *grid[2]={grid1,grid2};
    int edge[2];
    edge[0]=difference[0][0]<difference[0][1]?0:1;
    edge[1]=difference[1][0]<difference[1][1]?0:1;
    int match=difference[0][edge[0]]<difference[1][edge[1]]?0:1;
    
    edge[0]=(1-edge[0])*4;
    edge[1]=(1-edge[1])*4;
    match=(1-match)*4;
    
    int found=0;
    int i,j;
    int offset[4]={-2,-1,2,3};
    CvScalar combinedEdge;
    
    for (i=0;i<2 && found==0;i++)
    {
        combinedEdge=grid[1-i]->combinedEdge[orientation[1-i]+edge[1-i]];
        for (j=0;j<4;j++)
        {
            if (calculateDeltaE(combinedEdge, grid[i]->edge[(orientation[i]+offset[j]+8+edge[i]+match)%8])<DELTA_E_CONNECTED_EDGE_THRESHOLD)
            {
                found=1;
                break;
            }
        }
    }
    return found;
}

int varifyTarget(int minX,int minY,int maxX,int maxY,int gridCount,IplImage* object,int objectNumber,int imageNumber,int halfGridSize,IplImage* image)
{
    if (gridCount<MIN_GRID_COUNT)
    {
        return 0;
    }
    
    int xDim=maxX-minX;
    int yDim=maxY-minY;
    int minDim=5;
    int maxDim=20;
    int minArea=20;
    char* shapeName;
    
    if (xDim>maxDim || yDim>maxDim ||xDim<minDim || yDim<minDim || xDim>yDim*MAX_X_Y_RATIO || yDim>xDim*MAX_X_Y_RATIO)
    {
        return 0;
    }
    
    int found=0;
    int step       = object->widthStep/sizeof(uchar);
    uchar* data    = (uchar *)object->imageData;
    
    int i,j;
    IplImage* shape=cvCreateImage(cvSize(maxX-minX+3,maxY-minY+3), IPL_DEPTH_8U, 1);
    int shapeStep       = shape->widthStep/sizeof(uchar);
    uchar* shapeData    = (uchar *)shape->imageData;
    cvSet(shape, cvScalar(0, 0, 0, 0), NULL);
    
    for (i=minX;i<=maxX;i++)
    {
        for (j=minY; j<=maxY; j++) 
        {
            if (data[j*step+i]==objectNumber)
            {
                shapeData[(j-minY+1)*shapeStep+(i-minX+1)]=255;
            }
        }
    }
    
    CvSeq* contour = 0;
    CvMemStorage* contourStorage = cvCreateMemStorage(0);
    cvFindContours( shape, contourStorage, &contour, sizeof(CvContour),
                   CV_RETR_EXTERNAL, CV_CHAIN_APPROX_SIMPLE ,cvPoint(0, 0));
    
    for( ; contour != 0; contour = contour->h_next )
    {
        if (cvContourArea(contour, CV_WHOLE_SEQ, 0)>minArea)
        {
            shapeName=recognition(contour);
            if(shapeName!=NULL)
            {
                int imageMaxX=maxX*halfGridSize+halfGridSize;
                int imageMinX=minX*halfGridSize+halfGridSize;
                int imageMaxY=maxY*halfGridSize+halfGridSize;
                int imageMinY=minY*halfGridSize+halfGridSize;
                writeTargetInfo(imageNumber, imageMinX, imageMinY, imageMaxX, imageMaxY, shapeName);
                if (DRAW_RECT_AROUND_TARGET) {
                    cvRectangle(image, cvPoint(imageMinX, imageMinY), cvPoint(imageMaxX, imageMaxY), cvScalar(RECT_COLOR_B, RECT_COLOR_G, RECT_COLOR_R, 0), RECT_THICKNESS, 4, 0);
                }
                found=1;
                break;
            }
        }
    }
    
    cvReleaseImage(&shape);
    cvReleaseMemStorage(&contourStorage);
    
    return found;
}

void writeTargetInfo(int imageNumber,int minX,int minY,int maxX,int maxY,char* name)
{
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%d.%s",FILE_PATH_INFO,imageNumber,INFO_FORMAT);
    FILE* fp = fopen(path, "a");
    fprintf (fp, "Target: (%d,%d) %s\n",(maxX+minX)/2,(maxY+minY)/2,name);
    fclose(fp);
}

void freeGrid(GRID ***grid,int gridWidth,int gridHeight)
{
    int i,j;
    for( i=0;i<gridWidth;i++)
    {
        for (j=0; j<gridHeight; j++) 
        {
            free(grid[i][j]);
        }
        free(grid[i]);    
    }
    free(grid);
}

void filterBackgroundEdge(GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *gridQueue)
{
    int i;
    int x,y;
    int count=gridQueue->count;
    ARRAY_TWO_INT_QUEUE *buffer = NULL;
    int removed=0;
    
    for (i=0; i<count; i++)
    {
        removeFromArrayTwoIntQueue(gridQueue, &x, &y);
        if (grid[x][y]->potentialTarget==1)
        {
            if (x==0 || x==gridWidth-1 || y==0 || y==gridHeight-1)
            {
                if (buffer==NULL)
                {
                    buffer = mallocArrayTwoIntQueue(count);
                }
                if(!connectBackgroundGrid(x,y,grid, gridWidth, gridHeight, buffer))
                {
                    addToArrayTwoIntQueue(gridQueue, x, y);
                }
                else
                {
                    removed=1;
                }
            }
            else
            {
                addToArrayTwoIntQueue(gridQueue, x, y);
            }
        }
    }
    
    if (buffer!=NULL)
    {
        freeArrayTwoIntQueue(buffer);
        if (removed==1)
        {
            cleanupGridQueue(gridQueue, grid);
        }
    }
    
}

int connectBackgroundGrid(int x,int y,GRID ***grid,int gridWidth, int gridHeight,ARRAY_TWO_INT_QUEUE *buffer)
{    
    int xDirection[8]={-1,0,1,1,1,0,-1,-1};
    int yDirection[8]={-1,-1,-1,0,1,1,1,0};
    int currentX,currentY;
    int i;
    
    int onVEdge=(x==0 || x==gridWidth-1);
    int onHEdge=(y==0 || y==gridHeight-1);
    
    if (onHEdge && onVEdge) 
    {
        grid[x][y]->potentialTarget=0;
        addToArrayTwoIntQueue(buffer, x, y);
    }
    else 
    {
        int edgeNumber;
        if (onVEdge) {
            if (x==0)
            {
                edgeNumber=6;
            }
            else
            {
                edgeNumber=2;
            }
        }
        else
        {
            if (y==0)
            {
                edgeNumber=0;
            }
            else
            {
                edgeNumber=4;
            }
        }
        
        if(calculateDeltaE(grid[x+onHEdge][y+onVEdge]->combinedEdge[edgeNumber], grid[x-onHEdge][y-onVEdge]->combinedEdge[edgeNumber])>DELTA_E_EDGE_GRID_THRESHOLD)
        {
            grid[x][y]->potentialTarget=0;
            addToArrayTwoIntQueue(buffer, x, y);
        }
        else
        {
            return 0;
        }
    }
    
    while (removeFromArrayTwoIntQueue(buffer, &x, &y)) 
    {        
        for (i=0; i<8; i++) 
        {
            currentX=x+xDirection[i];
            currentY=y+yDirection[i];
            if (currentX>=0 && currentX<gridWidth && currentY>=0 && currentY<gridHeight)
            {
                if (grid[currentX][currentY]->potentialTarget==1 && connectedEdgeStrict(grid[x][y],grid[currentX][currentY]))
                {
                    grid[currentX][currentY]->potentialTarget=0;
                    addToArrayTwoIntQueue(buffer, currentX, currentY);
                }
            }
        }
    }
    return 1;
}

int connectedEdgeStrict(GRID *grid1,GRID *grid2)
{
    int orientation[2];
    orientation[0]=grid1->edgeOrientation;
    orientation[1]=grid2->edgeOrientation;
    float difference[2][2];
    
    difference[0][0]=calculateDeltaE(grid1->combinedEdge[orientation[0]], grid2->combinedEdge[orientation[0]]);
    difference[0][1]=calculateDeltaE(grid1->combinedEdge[orientation[1]+4], grid2->combinedEdge[orientation[1]+4]);
    if (difference[0][0]<DELTA_E_CONNECTED_EDGE_THRESHOLD && difference[0][1]<DELTA_E_CONNECTED_EDGE_THRESHOLD) 
    {
        return 1;
    }
    
    difference[1][0]=calculateDeltaE(grid1->combinedEdge[orientation[0]], grid2->combinedEdge[orientation[0]+4]);
    difference[1][1]=calculateDeltaE(grid1->combinedEdge[orientation[1]+4], grid2->combinedEdge[orientation[1]]);
    if (difference[1][0]<DELTA_E_CONNECTED_EDGE_THRESHOLD && difference[1][1]<DELTA_E_CONNECTED_EDGE_THRESHOLD)
    {
        return 1;
    }
    
    return 0;
}

void cleanupGridQueue(ARRAY_TWO_INT_QUEUE *gridQueue,GRID ***grid)
{
    int i;
    int count=gridQueue->count;
    int x,y;
    
    for (i=0; i<count; i++)
    {
        removeFromArrayTwoIntQueue(gridQueue, &x, &y);
        if (grid[x][y]->potentialTarget==1)
        {
            addToArrayTwoIntQueue(gridQueue, x, y);
        }
    }
}
