#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <semaphore.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>

#include "configuration.h"
#include "imageCapturing.h"
#include "imageProcessing.h"
#include "imageTransmission.h"
#include "ftpTransmission.h"
#include "commandTransmission.h"
#include "utility.h"
#include "camera.h"


void* imageCapturing(void* data);
void* imageProcessing(void* data);
void* imageTransmission(void* data);
void* commandTransmission(void* data);
void LaunchThread(void* thread);
void initializeData(void);
void addToCapturedImageBuffer(int number);
int removeFromCapturedImageBuffer(void);
void addToProcessedImageBuffer(int number,int targetCount);
int removeFromProcessedImageBuffer(void);
void initializeDirectories(void);
void deleteFile(const char* directory, const char *type,int imageNumber); 
int fileSize(const char* directory, const char *type,int imageNumber);
static void terminationSignal(int signalNum);
void setSignalHandler(void);
void cleanup(void);
void displayResult(void);
void displayStatus(void);
void initializeConfiguration(void);
void executeCommand(COMMAND *newCommand);
int getCompressionQuality(void);

sem_t* capturedImageEmpty;
sem_t* capturedImageFull;
sem_t* processedImageEmpty;
sem_t* processedImageFull;

int capturedImageEmptySemValue;
int processedImageEmptySemValue;

int processedImageCount;
int transmittedImageCount;
int capturedImageCount;
long int totalProcessTime;
long int totalTransmitTime;

long int processedFileSize;
long int transmittedFileSize;

int imageProcessingthreadCount;

int termination=0;
long int startTime;

int compressionQuality;

pthread_mutex_t captureMutex;
pthread_mutex_t processMutex;
pthread_mutex_t commandMutex;
pthread_mutex_t statusMutex;

ARRAY_INT_QUEUE* capturedImageBuffer;
ARRAY_INT_QUEUE* processedImageBuffer;
ARRAY_INT_QUEUE* targetImageBuffer;

COMMAND command;

void addToCapturedImageBuffer(int number)
{
    addToArrayIntQueueTail(capturedImageBuffer, number);
    if (capturedImageEmptySemValue==0 && CAPTURED_IMAGE_BUFFER_FULL==DROP_ON_FULL)
    {
        sem_wait(capturedImageFull);
        if (PROCESSING_AND_TRANSMISSION_PRIORITY==LEAST_RECENT)
        {
            removeFromArrayIntQueueTail(capturedImageBuffer);
        }
        else
        {
            removeFromArrayIntQueueHead(capturedImageBuffer);
        }
        capturedImageEmptySemValue++;
        sem_post(capturedImageEmpty);
    }
}

int removeFromCapturedImageBuffer(void)
{
    if (PROCESSING_AND_TRANSMISSION_PRIORITY==LEAST_RECENT)
    {
        return removeFromArrayIntQueueHead(capturedImageBuffer);
    }
    else
    {
        return removeFromArrayIntQueueTail(capturedImageBuffer);
    }
}
void addToProcessedImageBuffer(int number,int targetCount)
{
    if (targetCount>0 && PRIORITIZE_TARGET_TRSNSMISSION==ENABLED)
    {
        addToArrayIntQueueTail(targetImageBuffer, number);
    }
    else
    {
        addToArrayIntQueueTail(processedImageBuffer, number);
    }
    if (processedImageEmptySemValue==0 && PROCESSED_IMAGE_BUFFER_FULL==DROP_ON_FULL)
    {
        sem_wait(processedImageFull);
        ARRAY_INT_QUEUE* queue;
        if (processedImageBuffer->count>0)
        {
            queue=processedImageBuffer;
        }
        else
        {
            queue=targetImageBuffer;
        }
        
        if (PROCESSING_AND_TRANSMISSION_PRIORITY==LEAST_RECENT)
        {
            removeFromArrayIntQueueTail(queue);
        }
        else
        {
            removeFromArrayIntQueueHead(queue);
        }
        processedImageEmptySemValue++;
        sem_post(processedImageEmpty);
    }
}
int removeFromProcessedImageBuffer(void)
{
    ARRAY_INT_QUEUE* queue;
    if (targetImageBuffer->count>0 && PRIORITIZE_TARGET_TRSNSMISSION==ENABLED)
    {
        queue=targetImageBuffer;
    }
    else
    {
        queue=processedImageBuffer;
    }
    
    if (PROCESSING_AND_TRANSMISSION_PRIORITY==LEAST_RECENT)
    {
        return removeFromArrayIntQueueHead(queue);
    }
    else
    {
        return removeFromArrayIntQueueTail(queue);
    }
}

void* imageCapturing(void* data)
{
    int imageNumber;
    int shutterSpeed=command.value[CMD_SHUTTER_SPEED];
    int newShutterSpeed;
    initCamera(shutterSpeed);
    
    while(!termination) {
        sem_wait(capturedImageEmpty);
        
        pthread_mutex_lock(&captureMutex);
        capturedImageEmptySemValue--;
        pthread_mutex_unlock(&captureMutex);
        
        pthread_mutex_lock(&commandMutex);
        newShutterSpeed=command.value[CMD_SHUTTER_SPEED];
        pthread_mutex_unlock(&commandMutex);
        if (shutterSpeed!=newShutterSpeed)
        {
            shutterSpeed=newShutterSpeed;
            changeShutterSpeed(shutterSpeed);
        }
        
        imageNumber=captureControl();
        
        pthread_mutex_lock(&statusMutex);
        capturedImageCount++;
        pthread_mutex_unlock(&statusMutex);
        
        pthread_mutex_lock(&captureMutex);
        addToCapturedImageBuffer(imageNumber);
        pthread_mutex_unlock(&captureMutex);
        
        sem_post(capturedImageFull);
    }
    return NULL;
}

void* imageProcessing(void* data)
{
    int imageNumber;
    int targetCount=0;
    long int processStartTime;
    int size;
    while(!termination) {
        sem_wait(capturedImageFull);
        sem_wait(processedImageEmpty);
        
        pthread_mutex_lock(&processMutex);
        processedImageEmptySemValue--;
        pthread_mutex_unlock(&processMutex);
        
        pthread_mutex_lock(&captureMutex);
        imageNumber=removeFromCapturedImageBuffer();      
        pthread_mutex_unlock(&captureMutex);
        
        processStartTime=timer(0);
        targetCount=processControl(imageNumber,getCompressionQuality());
        
        size=fileSize(FILE_PATH_PROCESSED,PROCESSED_IMAGE_FORMAT,imageNumber);
        pthread_mutex_lock(&statusMutex);
        processedImageCount++;
        processedFileSize+=size;
        pthread_mutex_unlock(&statusMutex);
        
        pthread_mutex_lock(&processMutex);
        totalProcessTime+=timer(processStartTime)/1000;
        addToProcessedImageBuffer(imageNumber,targetCount);
        pthread_mutex_unlock(&processMutex);
        
        
        if (DELETE_IMAGE_AFTER_PROCESSING==ENABLED)
        {
            deleteFile(FILE_PATH_CAPTURED,CAPTURED_IMAGE_FORMAT,imageNumber);
        }
        
        pthread_mutex_lock(&captureMutex);
        capturedImageEmptySemValue++;
        pthread_mutex_unlock(&captureMutex);
        
        sem_post(processedImageFull);
        sem_post(capturedImageEmpty);
    }
    return NULL;
}

void* imageTransmission(void* data)
{
    initImageTransmission();
    int imageNumber;
    long int transmitStartTime;
    int size;
    while(!termination) {
        sem_wait(processedImageFull);
        
        pthread_mutex_lock(&processMutex);
        imageNumber=removeFromProcessedImageBuffer(); 
        pthread_mutex_unlock(&processMutex);
        
        transmitStartTime=timer(0);
        transmitControl(imageNumber);
        totalTransmitTime+=timer(transmitStartTime)/1000;
        
        size=fileSize(FILE_PATH_PROCESSED,PROCESSED_IMAGE_FORMAT,imageNumber);
        pthread_mutex_lock(&statusMutex);
        transmittedImageCount++;
        transmittedFileSize+=size;
        pthread_mutex_unlock(&statusMutex);
        
        if (DELETE_IMAGE_AFTER_TRANSMISSION==ENABLED)
        {
            deleteFile(FILE_PATH_PROCESSED,PROCESSED_IMAGE_FORMAT,imageNumber);
        }
        if (DELETE_INFO_AFTER_TRANSMISSION==ENABLED)
        {
            deleteFile(FILE_PATH_INFO,INFO_FORMAT,imageNumber);
        }
        
        pthread_mutex_lock(&processMutex);
        processedImageEmptySemValue++;
        pthread_mutex_unlock(&processMutex);
        
        sem_post(processedImageEmpty);
    }
    return NULL;
}

void* commandTransmission(void* data)
{
    COMMAND newCommand=command;
    while(!termination) 
    {
        commandTransmissionControl(command,&newCommand);
        executeCommand(&newCommand);
    }
    return NULL;
}

void  executeCommand(COMMAND *newCommand)
{
    int i;
    for (i=0; i<NUMBER_OF_CONNANDS; i++) 
    {
        if (command.value[i]!=newCommand->value[i])
        {
            pthread_mutex_lock(&commandMutex);
            command.value[i]=newCommand->value[i];
            pthread_mutex_unlock(&commandMutex);
        }
    }
}

int getCompressionQuality(void)
{
    static long int processBufferSize=0;
    
    pthread_mutex_lock(&commandMutex);
    int newCompressionQuality=compressionQuality;
    pthread_mutex_unlock(&commandMutex);
    
    pthread_mutex_lock(&statusMutex);
    long int newProcessBufferSize=processedFileSize-transmittedFileSize;
    long int oldProcessBufferSize=processBufferSize;
    processBufferSize=newProcessBufferSize;
    pthread_mutex_unlock(&statusMutex);
    
    pthread_mutex_lock(&captureMutex);
    int captureBufferCount=capturedImageBuffer->count;
    pthread_mutex_unlock(&captureMutex);
    
    pthread_mutex_lock(&processMutex);
    int processBufferCount=processedImageBuffer->count;
    pthread_mutex_unlock(&processMutex);
        
    if (newProcessBufferSize>oldProcessBufferSize || processBufferCount>=PROCESSED_IMAGE_BUFFER_SIZE-2) {
        newCompressionQuality-=COMPRESSION_QUALITY_CHANGE_FOR_TRANSMISSION;
        if (newCompressionQuality<MIN_COMPRESSION_QUALITY_FOR_OTHER_IMAGE) {
            newCompressionQuality=MIN_COMPRESSION_QUALITY_FOR_OTHER_IMAGE;
        }
    }
    else if (processBufferCount<=PROCESSED_IMAGE_BUFFER_SIZE/2)
    {
        newCompressionQuality+=COMPRESSION_QUALITY_CHANGE_FOR_TRANSMISSION;
        if (newCompressionQuality>MAX_COMPRESSION_QUALITY) {
            newCompressionQuality=MAX_COMPRESSION_QUALITY;
        }
    }
    
    if (captureBufferCount>=CAPTURED_IMAGE_BUFFER_SIZE-imageProcessingthreadCount-1)
    {
        newCompressionQuality-=COMPRESSION_QUALITY_CHANGE_FOR_CAPTURING;
        if (newCompressionQuality<MIN_COMPRESSION_QUALITY_FOR_OTHER_IMAGE) {
            newCompressionQuality=MIN_COMPRESSION_QUALITY_FOR_OTHER_IMAGE;
        }
    }
    
    pthread_mutex_lock(&commandMutex);
    compressionQuality=newCompressionQuality;
    pthread_mutex_unlock(&commandMutex);
    
    return newCompressionQuality;
}

void LaunchThread(void* thread)
{
    // Create the thread using POSIX routines.
    pthread_attr_t  attr;
    pthread_t       posixThreadID;
    int             returnVal;
    
    returnVal = pthread_attr_init(&attr);
    assert(!returnVal);
    returnVal = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    assert(!returnVal);
    
    int     threadError = pthread_create(&posixThreadID, &attr, thread, NULL);
    
    returnVal = pthread_attr_destroy(&attr);
    assert(!returnVal);
    if (threadError != 0)
    {
        // Report an error.
    }
}

void initializeData(void)
{    
    pthread_mutex_init(&captureMutex, NULL);
    sem_unlink("capturedImageFull");
    sem_unlink("capturedImageEmpty");
    capturedImageFull=sem_open("capturedImageFull",O_CREAT, 0, 0);
    capturedImageEmpty=sem_open("capturedImageEmpty", O_CREAT,0, CAPTURED_IMAGE_BUFFER_SIZE);
    capturedImageBuffer=mallocArrayIntQueue(CAPTURED_IMAGE_BUFFER_SIZE);
    capturedImageEmptySemValue=CAPTURED_IMAGE_BUFFER_SIZE;
    
    pthread_mutex_init(&processMutex, NULL);
    sem_unlink("processedImageFull");
    sem_unlink("processedImageEmpty");
    processedImageFull=sem_open("processedImageFull",O_CREAT, 0, 0);
    processedImageEmpty=sem_open("processedImageEmpty", O_CREAT,0, PROCESSED_IMAGE_BUFFER_SIZE);
    processedImageBuffer=mallocArrayIntQueue(PROCESSED_IMAGE_BUFFER_SIZE);
    targetImageBuffer=mallocArrayIntQueue(PROCESSED_IMAGE_BUFFER_SIZE);
    processedImageEmptySemValue=PROCESSED_IMAGE_BUFFER_SIZE;
    
    pthread_mutex_init(&commandMutex, NULL);
    pthread_mutex_init(&statusMutex, NULL);
    
    processedImageCount=0;
    transmittedImageCount=0;
    totalProcessTime=1;
    totalTransmitTime=1;
    transmittedFileSize=0;
    processedFileSize=0;
    
    compressionQuality=(MIN_COMPRESSION_QUALITY_FOR_OTHER_IMAGE+MAX_COMPRESSION_QUALITY)/2;
    
    imageProcessingthreadCount=(IMAGE_PROCESSING_THREAD_COUNT==DEFAULT)?(int)sysconf(_SC_NPROCESSORS_ONLN):IMAGE_PROCESSING_THREAD_COUNT;
}

void  initializeConfiguration(void)
{
    initCommandTransmission();
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%s",LOCALPATH_COMMAND,COMMAND_FILENAME);
    char url[URL_LENGTH];
    sprintf(url, "%s/%s",GROUND_STATION_URL_COMMAND,COMMAND_FILENAME);
    
    command.value[CMD_SHUTTER_SPEED]=SHUTTER_SPEED;
    command.value[CMD_TEST]=1;
    
    transmitCommand(url, path);
    COMMAND oldCommand=command;
    loadCommand(path,oldCommand,&command);
}

int main (int argc, const char * argv[])
{    
      long int time=timer(0);
      for (int i=51;i<60;i++)
      {
      processControl(i, 50);
      }
      printf("%.2f\n",9.0/(float)timer(time)*1000000);
      
    /*
    initializeData();
    initializeDirectories();
    initializeConfiguration();
    
    long int lastTime=timer(0);
    startTime=timer(0);
    int i;
    
    LaunchThread(commandTransmission);
    LaunchThread(&imageCapturing);
    for (i=0;i<imageProcessingthreadCount;i++)
    {
        LaunchThread(&imageProcessing);
    }
    initTransmission();
    LaunchThread(&imageTransmission);
    setSignalHandler();
    
    while (1)
    {
        sleep(FPS_REPORT_FREQUENCY);
        if (timer(lastTime)>=FPS_REPORT_FREQUENCY*1000000)
        {
            lastTime=timer(0);
            displayStatus();
        }
    }*/
}

void displayStatus(void)
{
    printf("Capture |Total: %d|FPS: %.2f\n",capturedImageCount,(float)capturedImageCount/((float)timer(startTime)/1000000.0));
    printf("Buffer %d\n",CAPTURED_IMAGE_BUFFER_SIZE-capturedImageEmptySemValue);
    printf("Process |Total: %d|FPS: %.2f|Producting Rate:   %.2f MB/s\n",processedImageCount,(float)processedImageCount/((float)totalProcessTime/1000.0)*imageProcessingthreadCount,(float)processedFileSize/(float)timer(startTime));
    printf("Buffer %d\n",PROCESSED_IMAGE_BUFFER_SIZE-processedImageEmptySemValue);
    printf("Transmit|Total: %d|FPS: %.2f|Transmitting Rate: %.2f MB/s\n",transmittedImageCount,(float)transmittedImageCount/(totalTransmitTime/1000.0),(float)transmittedFileSize/(float)timer(startTime));
    printf("\n");
}

void initializeDirectories(void)
{
    mkdir(FILE_PATH_CAPTURED, 0777);
    mkdir(FILE_PATH_PROCESSED, 0777);
    mkdir(FILE_PATH_INFO, 0777);
    mkdir(LOCALPATH_COMMAND, 0777);
}

void deleteFile(const char* directory, const char *type,int imageNumber)
{
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%d.%s",directory,imageNumber,type);
    remove(path);
}

int fileSize(const char* directory, const char *type,int imageNumber)
{
    char path[FILE_PATH_LENGTH];
    sprintf(path, "%s/%d.%s",directory,imageNumber,type);
    struct stat st;
    stat(path, &st);
    return (int)st.st_size;
}

void setSignalHandler(void)
{
    struct sigaction* newAction;
    newAction=malloc(sizeof(struct sigaction));
    newAction->sa_handler = (terminationSignal);
    sigaction(SIGINT, newAction, NULL);
}

static void terminationSignal(int signalNum)
{
    if (signalNum==SIGINT)
    {
        termination=1;
        printf("\nTermination Signal Received. Closing in %d seconds.\n",DELAY_BEFORE_TERMINATION);
        sleep(DELAY_BEFORE_TERMINATION);
        cleanup();
        displayResult();
        exit(0);
    }
}

void displayResult(void)
{
    printf("Total time:%d:%d\n",(int)(timer(startTime)/1000000/60),(int)(timer(startTime)/1000000)%60);
    displayStatus();
}

void cleanup(void)
{
    freeArrayIntQueue(capturedImageBuffer);
    freeArrayIntQueue(processedImageBuffer);
    freeArrayIntQueue(targetImageBuffer);  
    cleanupCamera();
    cleanupCommandTransmission();
    cleanupImageTransmission();
    cleanupTransmission();
}